# coding: utf-8

"""
Прямой доступ к базе данных.
"""

from io import SEEK_SET
from ctypes import BigEndianStructure, c_int32, c_uint16, c_uint32
from typing import TYPE_CHECKING
from cbsvibpyirbis._common import change_extension, UTF
from cbsvibpyirbis.records import Field, RawRecord, Record
if TYPE_CHECKING:
    from typing import List

#############################################################################


class XrfRecord(BigEndianStructure):
    """
    Содержит информацию о смещении и статусе записи в мастер-файле.
    """
    _fields_ = [('offset_low', c_uint32),
                ('offset_high', c_uint32),
                ('status', c_int32)]

    def offset(self) -> int:
        """
        Вычисление полного смещения для данной записи.
        :return:
        """
        return self.offset_low + (self.offset_high << 32)

    def __str__(self):
        return f"{self.offset()}: {self.status}"


#############################################################################


class XrfFile:
    """
    Файл перекрестных ссылок.
    """

    __slots__ = ('_stream',)

    def __init__(self, filename: str) -> None:
        self._stream = open(filename, 'rb')

    @staticmethod
    def _get_offset(mfn: int) -> int:
        return (mfn - 1) * 12

    def close(self) -> None:
        """
        Закрывает файл.
        :return: None.
        """
        self._stream.close()

    def read_record(self, mfn: int) -> XrfRecord:
        """
        Чтение одной записи.
        :param mfn:
        :return:
        """
        offset = self._get_offset(mfn)
        self._stream.seek(offset, SEEK_SET)
        result = XrfRecord()
        self._stream.readinto(result)  # type: ignore
        return result

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return exc_type is None


#############################################################################

class MstLeader(BigEndianStructure):
    """
    Лидер записи.
    """

    _fields_ = [('mfn', c_uint32),
                ('length', c_uint32),
                ('back_low', c_uint32),
                ('back_high', c_uint32),
                ('base', c_uint32),
                ('nvf', c_uint32),
                ('version', c_uint32),
                ('status', c_uint32)]

    def __str__(self):
        return f"MFN {self.mfn}, version {self.version}, status {self.status}"


#############################################################################


class MstEntry(BigEndianStructure):
    """
    Элемент справочника MST-файла,
    описывающий поле переменной длины.
    """

    _fields_ = [('tag', c_uint32),
                ('position', c_uint32),
                ('length', c_uint32)]

    def __str__(self):
        return f"{self.tag}: {self.position}: {self.length}"


#############################################################################


class MstControl(BigEndianStructure):
    """
    Первая запись в файле документов – управляющая
    запись, которая формируется (в момент определения
    базы данных или при ее инициализации) и поддерживается
    автоматически.
    """

    _fields_ = [('ctl_mfn', c_uint32),
                ('next_mfn', c_uint32),
                ('next_low', c_uint32),
                ('next_high', c_uint32),
                ('mft_type', c_uint32),
                ('rec_cnt', c_uint32),
                ('mfc_xx1', c_uint32),
                ('mfc_xx2', c_uint32),
                ('mfc_xx3', c_uint32)]

    def next_offset(self) -> int:
        """
        Смещение на свободное место.
        :return:
        """
        return self.next_low + (self.next_high << 32)

    def __str__(self):
        return f"{self.next_mfn}: {self.next_offset()}"

#############################################################################


class MstField:
    """
    Поле в мастер-записи.
    """

    __slots__ = 'tag', 'value'

    def __init__(self):
        self.tag: int = 0
        self.value: str = ''

    def __str__(self):
        return f"{self.tag}: {self.value}"


#############################################################################


class MstRecord:
    """
    Запись в мастер-файле.
    """

    __slots__ = 'leader', 'fields'

    def __init__(self):
        self.leader: MstLeader = MstLeader()
        self.fields: 'List[MstField]' = []

    def decode_raw(self) -> RawRecord:
        """
        Декодирование в сырую запись.
        :return:
        """
        result = RawRecord()
        result.mfn = self.leader.mfn
        result.version = self.leader.version
        result.status = self.leader.status
        for field in self.fields:
            line = f"{field.tag}#{field.value}"
            result.fields.append(line)
        return result

    def decode_record(self) -> Record:
        """
        Декодирование в полноценную запись.
        :return:
        """
        result = Record()
        result.mfn = self.leader.mfn
        result.version = self.leader.version
        result.status = self.leader.status
        for source in self.fields:
            target = Field(source.tag)
            target.headless_parse(source.value)
            result.fields.append(target)
        return result

    def dump_fields(self) -> None:
        """
        Дамп полей на консоль.
        :return: None.
        """
        for field in self.fields:
            print(field)
        print('-' * 70)

    def __str__(self):
        return str(self.leader)


#############################################################################


class MstFile:
    """
    Мастер-файл базы данных.
    """

    __slots__ = '_stream', 'control', '_filename'

    def __init__(self, filename: str) -> None:
        """
        Открытие файла.
        :param filename: Имя файла.
        """
        self._filename = filename
        self._stream = open(filename, 'rb')
        self.control = MstControl()
        self._stream.readinto(self.control)  # type: ignore

    def close(self) -> None:
        """
        Закрывает файл.
        :return:
        """
        self._stream.close()

    def next_mfn(self) -> int:
        """
        Получение следующего MFN.
        :return:
        """
        return self.control.next_mfn

    def read_record(self, offset: int) -> MstRecord:
        """
        Чтение записи по указанному смещению.
        :param offset: Смещение.
        :return: Прочитанная запись.
        """
        self._stream.seek(offset, SEEK_SET)
        result = MstRecord()
        self._stream.readinto(result.leader)  # type: ignore
        nvf = result.leader.nvf
        entries: 'List[MstEntry]' = []
        for i in range(nvf):
            entry = MstEntry()
            self._stream.readinto(entry)  # type: ignore
            entries.append(entry)
        base = result.leader.base
        for i in range(nvf):
            field = MstField()
            field.tag = entries[i].tag
            position = entries[i].position
            self._stream.seek(offset + base + position, SEEK_SET)
            field_length = entries[i].length
            raw_value = self._stream.read(field_length)
            field.value = str(raw_value, UTF)
            result.fields.append(field)
        return result

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return exc_type is None

    def __str__(self):
        return f"{self._filename}: {self.next_mfn()}"


#############################################################################


class NodeItem(BigEndianStructure):
    """
    Элемент справочника в N01/L01.
    """

    _fields_ = [('length', c_uint16),
                ('key_offset', c_uint16),
                ('offset_low', c_int32),
                ('offset_high', c_int32)]

    def offset(self):
        """
        Полное смещение.
        :return:
        """
        return self.offset_high + (self.offset_low << 32)

    def __str__(self):
        return f"{self.length}: {self.key_offset}"


#############################################################################


class NodeLeader(BigEndianStructure):
    """
    Лидер записи в L01/N01.
    """

    _fields_ = [('number', c_uint32),
                ('previous', c_int32),
                ('next', c_uint32),
                ('term_count', c_int32),
                ('free_offset', c_int32)]

    def __str__(self):
        return f"{self.number}"


#############################################################################


class NodeRecord:
    """
    Файлы N01 и L01 содержат  в себе индексы словаря поисковых
    терминов и состоят из записей (блоков) постоянной длины.
    Записи состоят из трех частей: лидера, справочника
    и ключей переменной длины.
    """

    __slots__ = 'leader', 'items', 'keys'

    def __init__(self):
        self.leader: NodeLeader = NodeLeader()
        self.items: 'List[NodeItem]' = []
        self.keys: 'List[str]' = []

    def __str__(self):
        return str(self.leader)


#############################################################################


class Link(BigEndianStructure):
    """
    Ссылка на запись в мастер-файле.
    """

    _fields_ = [('mfn', c_uint32),
                ('tag', c_int32),
                ('occ', c_int32),
                ('cnt', c_int32)]

    def __str__(self):
        return f"{self.mfn}: {self.tag}: {self.occ}: {self.cnt}"


#############################################################################


class InvertedLeader(BigEndianStructure):
    """
    Заголовок инвертированной записи.
    """

    _fields_ = [('low', c_int32),
                ('high', c_int32),
                ('total', c_int32),
                ('used', c_int32),
                ('count', c_int32)]

    def __str__(self):
        return f"{self.high}: {self.low}"


#############################################################################


class InvertedRecord:
    """
    Запись состоит из заголовка и упорядоченного набора ссылок.
    """

    __slots__ = 'leader', 'links'

    def __init__(self):
        self.leader: InvertedLeader = InvertedLeader()
        self.links: 'List[Link]' = []

    def __str__(self):
        return str(self.leader)


#############################################################################


class InvertedFile:
    """
    Работа с инверсным файлом.
    """

    NODE_LENGTH = 2048

    __slots__ = '_ifp', '_l01', '_n01'

    def __init__(self, filename: str) -> None:
        self._ifp = open(change_extension(filename, '.ifp'), 'rb')
        self._l01 = open(change_extension(filename, '.l01'), 'rb')
        self._n01 = open(change_extension(filename, '.n01'), 'rb')

    def close(self):
        """
        Закрывает файлы.
        :return:
        """
        self._ifp.close()
        self._l01.close()
        self._n01.close()

    @staticmethod
    def _node_offset(number: int) -> int:
        return (number - 1) * InvertedFile.NODE_LENGTH

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return exc_type is None

#############################################################################


class DirectAccess:
    """
    Прямой доступ к базе данных.
    """

    __slots__ = '_xrf', '_mst', '_inverted', '_filename'

    def __init__(self, master_file_name: str) -> None:
        self._filename = master_file_name
        self._mst = MstFile(master_file_name)
        xrf_file_name = change_extension(master_file_name, '.xrf')
        self._xrf = XrfFile(xrf_file_name)
        self._inverted = InvertedFile(master_file_name)

    def close(self) -> None:
        """
        Закрывает файлы.
        :return:
        """
        self._mst.close()
        self._xrf.close()
        self._inverted.close()

    def next_mfn(self) -> int:
        """
        Получает следующий MFN.
        :return:
        """
        return self._mst.next_mfn()

    def read_mst_record(self, mfn: int) -> MstRecord:
        """
        Чтение MST-записи.
        :param mfn: MFN.
        :return: Прочитанная запись.
        """
        xrf_record = self._xrf.read_record(mfn)
        result = self._mst.read_record(xrf_record.offset())
        return result

    def read_raw_record(self, mfn: int) -> RawRecord:
        """
        Чтение RAW-записи.
        :param mfn: MFN.
        :return: Прочитанная запись.
        """
        mst_record = self.read_mst_record(mfn)
        return mst_record.decode_raw()

    def read_record(self, mfn: int) -> Record:
        """
        Чтение полностью декодированной записи.
        :param mfn: MFN.
        :return:
        """
        mst_record = self.read_mst_record(mfn)
        return mst_record.decode_record()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return exc_type is None

    def __str__(self):
        return f"{self._filename}: {self.next_mfn()}"


#############################################################################


__all__ = ['DirectAccess', 'InvertedFile', 'MstControl', 'MstField', 'MstFile',
           'MstEntry', 'MstRecord', 'XrfFile', 'XrfRecord']



# coding: utf-8

"""
Версия сервера.
"""

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import List


class ServerVersion:
    """
    Информация о версии сервера, числе подключенных/разрешенных клиентов
    и организации, на которую зарегистрирован сервер.
    """

    def __init__(self):
        self.organization = ''
        self.version = ''
        self.max_clients = 0
        self.connected_clients = 0

    def parse(self, lines: 'List[str]') -> None:
        """
        Parse the text.

        :param lines: Text to parse
        :return: None
        """
        if len(lines) == 3:
            self.version = lines[0]
            self.connected_clients = int(lines[1])
            self.max_clients = int(lines[2])
        else:
            self.organization = lines[0]
            self.version = lines[1]
            self.connected_clients = int(lines[2])
            self.max_clients = int(lines[3])

    def __str__(self):
        buffer = [self.organization,
                  self.version,
                  str(self.connected_clients),
                  str(self.max_clients)]
        return '\n'.join(buffer)


__all__ = ['ServerVersion']



# coding: utf-8

"""
Константы и утилитные функции.
"""

import asyncio
import os.path
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Union, Optional, SupportsInt, List

#############################################################################
# COMMON CONSTANTS
#############################################################################

# Max number of postings

MAX_POSTINGS = 32758

# Encodings

ANSI = 'cp1251'
UTF = 'utf-8'
OEM = 'cp866'

# Stop marker for menus

STOP_MARKER = '*****'

# Record status

LOGICALLY_DELETED = 1
PHYSICALLY_DELETED = 2
ABSENT = 4
NON_ACTUALIZED = 8
NEW_RECORD = 16
LAST = 32
LOCKED = 64
AUTOIN_ERROR = 128
FULL_TEXT_NOT_ACTUALIZED = 256

# Paths

SYSTEM = 0
DATA = 1
MASTER_FILE = 2
INVERTED_FILE = 3
PARAMETER_FILE = 10
FULL_TEXT = 11
INTERNAL_RESOURCE = 12

# Text delimiters

IRBIS_DELIMITER = '\x1F\x1E'
SHORT_DELIMITER = '\x1E'
OTHER_DELIMITER = '\x1F'
MSDOS_DELIMITER = '\r\n'

# Common formats

ALL = "&uf('+0')"
BRIEF = '@brief'
IBIS = '@ibiskw_h'
INFORMATIONAL = '@info_w'
OPTIMIZED = '@'

# Workstation codes

ADMINISTRATOR = 'A'
CATALOGER = 'C'
ACQUISITIONS = 'M'
COMPLECT = 'M'
READER = 'R'
CIRCULATION = 'B'
BOOKLAND = 'B'
PROVISION = 'K'
JAVA_APPLET = 'J'

# Error codes

# Клиент не подключен к серверу
NOT_CONNECTED = -100500

# Command codes

# Получение признака монопольной блокировки базы данных
EXCLUSIVE_DATABASE_LOCK = '#'

# Получение списка удаленных, неактуализированных
# и заблокированных записей
RECORD_LIST = '0'

# Получение версии сервера
SERVER_INFO = '1'

# Получение статистики по базе данных
DATABASE_STAT = '2'

# Глобальная корректировка
GLOBAL_CORRECTION = '5'

# Сохранение группы записей
SAVE_RECORD_GROUP = '6'

# Печать
PRINT = '7'

# Запись параметров в ini - файл, расположенный на сервере
UPDATE_INI_FILE = '8'

IMPORT_ISO = '9'

# Регистрация клиента на сервере
REGISTER_CLIENT = 'A'

# Разрегистрация клиента
UNREGISTER_CLIENT = 'B'

# Чтение записи, ее расформатирование
READ_RECORD = 'C'

# Сохранение записи
UPDATE_RECORD = 'D'

# Разблокировка записи
UNLOCK_RECORD = 'E'

# Актуализация записи
ACTUALIZE_RECORD = 'F'

# Форматирование записи или группы записей
FORMAT_RECORD = 'G'

# Получение терминов и ссылок словаря,
# форматирование записей
READ_TERMS = 'H'

# Получение ссылок для термина (списка терминов)
READ_POSTINGS = 'I'

# Глобальная корректировка виртуальной записи
CORRECT_VIRTUAL_RECORD = 'J'

# Поиск записей с опциональным форматированием
SEARCH = 'K'

# Получение / сохранение текстового файла,
# расположенного на сервере (группы текстовых файлов)
READ_DOCUMENT = 'L'

BACKUP = 'M'

# Пустая операция. Периодическое подтверждение
# соединения с сервером
NOP = 'N'

# Получение максимального MFN для базы данных
GET_MAX_MFN = 'O'

# Получение терминов и ссылок словаря в обратном порядке
READ_TERMS_REVERSE = 'P'

# Разблокирование записей
UNLOCK_RECORDS = 'Q'

# Полнотекстовый поиск (новый!)
FULL_TEXT_SEARCH = '&'

# Опустошение базы данных
EMPTY_DATABASE = 'S'

# Создание базы данных
CREATE_DATABASE = 'T'

# Разблокирование базы данных
UNLOCK_DATABASE = 'U'

# Чтение ссылок для заданного MFN
GET_RECORD_POSTINGS = 'V'

# Удаление базы данных
DELETE_DATABASE = 'W'

# Реорганизация мастер - файла
RELOAD_MASTER_FILE = 'X'

# Реорганизация словаря
RELOAD_DICTIONARY = 'Y'

# Создание поискового словаря заново
CREATE_DICTIONARY = 'Z'

# Получение статистики работы сервера
GET_SERVER_STAT = '+1'

# Получение списка запущенных процессов
GET_PROCESS_LIST = '+3'

# Сохранение списка пользователей
SET_USER_LIST = '+7'

# Перезапуск сервера
RESTART_SERVER = '+8'

# Получение списка пользователей
GET_USER_LIST = '+9'

# Получение списка файлов на сервере
LIST_FILES = '!'

# Codes for ReadRecord command
READ_RECORD_CODES = [-201, -600, -602, -603]

# Codes for ReadTerms command
READ_TERMS_CODES = [-202, -203, -204]

#############################################################################
# COMMON FUNCTIONS
#############################################################################


def same_string(first: 'Optional[str]', second: 'Optional[str]') -> bool:
    """
    Строки совпадают с точностью до регистра?

    :param first: Первая строка
    :param second: Вторая строка
    :return: True, если совпадают
    """
    if first is None or second is None:
        return False
    return first.upper() == second.upper()


def safe_str(obj) -> str:
    """
    Простое избегание 'None' при выводе текстового представления объекта.

    :param obj: Объект
    :return: Текстовое представление
    """
    if obj is None:
        return ''
    return str(obj)


def safe_int(text: 'Union[str, bytes, SupportsInt]') -> int:
    """
    Безопасное превращение строки в целое.

    :param text: Текст
    :return: Целое число
    """
    try:
        result: int = int(text)
    except ValueError:
        result = 0
    return result


# noinspection PyUnreachableCode
def throw_value_error() -> str:
    """
    Simple raises ValueError.

    :return: None
    """
    raise ValueError()


def irbis_to_dos(text: str) -> str:
    """
    Convert IRBIS text to MS-DOS text.

    :param text: Text to convert
    :return: Converted text
    """
    return text.replace(IRBIS_DELIMITER, '\n')


def irbis_to_lines(text: str) -> 'List[str]':
    """
    Convert IRBIS text to list of strings.

    :param text: Text to convert
    :return: List of strings
    """
    return text.split(IRBIS_DELIMITER)


def short_irbis_to_lines(text: str) -> 'List[str]':
    """
    Convert IRBIS with short delimiter text to list of strings.

    :param text: Text to convert
    :return: List of strings
    """
    return text.split(SHORT_DELIMITER)


def change_extension(filename: str, new_extension: str) -> str:
    """
    Смена расширения у файла (без физического переименования!).
    :param filename:
    :param new_extension:
    :return:
    """
    return os.path.splitext(filename)[0] + new_extension


def remove_comments(text: str) -> str:
    """
    Удаление комментариев из кода.

    :param text: Код для очистки
    :return: Очищенный код
    """

    if '/*' not in text:
        return text

    result = []
    state = ''
    index = 0
    length = len(text)
    while index < length:
        char = text[index]
        if state in ("'", '"', '|'):
            if char == state:
                state = ''
            result.append(char)
        else:
            if char == '/':
                if index + 1 < length and text[index + 1] == '*':
                    while index < length:
                        char = text[index]
                        if char in ('\r', '\n'):
                            result.append(char)
                            break
                        index = index + 1
                else:
                    result.append(char)
            else:
                if char in ("'", '"', '|'):
                    state = char
                    result.append(char)
                else:
                    result.append(char)
        index = index + 1

    return ''.join(result)


def prepare_format(text: str) -> str:
    """
    Подготовка формата к отсылке на сервер.

    :param text: Неподготовленный формат
    :return: Подготовленный формат
    """
    text = remove_comments(text)
    length = len(text)
    if length == 0:
        return text

    flag = False
    for char in text:
        if char < ' ':
            flag = True
            break

    if not flag:
        return text

    result = []
    for char in text:
        if char >= ' ':
            result.append(char)

    return ''.join(result)


##############################################################################

# pylint: disable=invalid-name
irbis_event_loop = None


def init_async() -> None:
    """
    Инициализация асинхронной обработки цикла сообщений.

    :return: None.
    """
    global irbis_event_loop  # pylint: disable=global-statement
    if not irbis_event_loop:
        try:
            irbis_event_loop = asyncio.get_running_loop()
        except RuntimeError:
            irbis_event_loop = asyncio.new_event_loop()

def get_event_loop() -> None | asyncio.AbstractEventLoop:
    """
    Получение асинхронного цикла сообщений.

    :return None | asyncio.AbstractEventLoop
    """
    global irbis_event_loop
    return irbis_event_loop

def close_async() -> None:
    """
    Завершение асинхронной обработки цикла сообщений.

    :return: None.
    """
    global irbis_event_loop  # pylint: disable=global-statement
    if irbis_event_loop:
        irbis_event_loop.close()
        irbis_event_loop = None


##############################################################################


class ObjectWithError:
    """
    Объект, хранящий код последней ошибки.
    """

    __slots__ = ('last_error',)

    def __init__(self):
        self.last_error: int = 0



# coding: utf-8

"""
Ответ сервера.
"""

import socket
from typing import TYPE_CHECKING
from cbsvibpyirbis._common import ANSI, ObjectWithError, UTF

if TYPE_CHECKING:
    from typing import Any, List, Optional


class ServerResponse:
    """
    Ответ сервера.
    """

    __slots__ = (
        "_memory",
        "_view",
        "_pos",
        "command",
        "client_id",
        "query_id",
        "length",
        "version",
        "return_code",
        "_conn",
    )

    def __init__(self, conn: ObjectWithError) -> None:
        self._conn: ObjectWithError = conn
        self._memory: bytearray = bytearray()
        self._view: memoryview = memoryview(bytearray())
        self._pos: int = 0  # Текущая позиция при чтении
        self.command: str = ""
        self.client_id: int = 0
        self.query_id: int = 0
        self.length: int = 0
        self.version: str = ""
        self.return_code: int = 0

    def read_data(self, sock: socket.socket) -> None:
        """
        Считывание ответа сервера из сокета.

        :param sock: Сокет для чтения.
        :return: None.
        """
        while sock:
            buffer = sock.recv(4096)
            if not buffer:
                break
            self._memory.extend(buffer)
        sock.close()
        self._view = memoryview(self._memory)

    async def read_data_async(self, sock: "Any") -> None:
        """
        Асинхронное считывание ответа из сокета.

        :param sock: Сокет для чтения.
        :return: None.
        """
        while True:
            buffer = await sock.read(4096)
            if not buffer:
                break
            self._memory.extend(buffer)
        self._view = memoryview(self._memory)

    def initial_parse(self) -> None:
        """
        Разбор постоянной начальной части ответа от сервера.

        :return: None.
        """
        self.command = self.ansi()
        self.client_id = self.number()
        self.query_id = self.number()
        self.length = self.may_be_number()
        self.version = self.ansi()
        self.return_code = 0
        for _ in range(5):
            self.read()

    def ansi(self) -> str:
        """
        Считывание одной строки в кодировке ANSI.

        :return: Строка (возможно, пустая)
        """
        # noinspection PyTypeChecker
        return str(self.read(), ANSI)  # type: ignore

    def ansi_n(self, count: int) -> "List[str]":
        """
        Считывание не менее указанного количества строк
        в кодировке ANSI.

        :param count: Сколько строк надо
        :return: Список строк или None
        """
        result = []
        for _ in range(count):
            line = self.ansi()
            if not line:
                return []
            result.append(line)
        return result

    def ansi_remaining_text(self) -> str:
        """
        Получение всего оставшегося текста ответа сервера
        как строки в кодировке ANSI.

        :return: Строка (возможно, пустая)
        """
        # noinspection PyTypeChecker
        return str(self._view[self._pos :], ANSI)  # type: ignore

    def ansi_remaining_lines(self) -> "List[str]":
        """
        Получение всего оставшегося текста ответа сервера
        как списка строк в кодировке ANSI.

        :return: Список строк (возможно, пустой)
        """
        result = []
        while True:
            line = self.ansi()
            if not line:
                break
            result.append(line)
        return result

    def check_return_code(self, allowed: "Optional[List[int]]" = None) -> bool:
        """
        Проверка кода возврата. Если код меньше нуля,
        генерируется IrbisError.
        Можно указать допустимые значения кода.

        :param allowed: Допустимые отрицательные значения (опционально).
        :return: Результат проверки кода возврата.
        """
        if allowed is None:
            allowed = []

        if self.get_return_code() < 0:
            if self.return_code not in allowed:
                #  raise IrbisError(self.return_code)
                return False

        return True

    def close(self) -> None:
        """
        Закрытие сокета.

        :return: None
        """

        # Пока ничего не делаем

    def get_binary_file(self) -> "Optional[bytearray]":
        """
        Получение двоичного файла с сервера.

        :return: Содержимое файла или None
        """

        preamble = bytearray(b"IRBIS_BINARY_DATA")
        index = self._memory.find(preamble)
        if index < 0:
            return None
        return self._memory[index + len(preamble) :]

    def get_return_code(self) -> int:
        """
        Считывание кода возврата без его проверки.

        :return: Код возврата
        """

        self.return_code = self.number()
        if self.return_code < 0:
            self._conn.last_error = self.return_code
        return self.return_code

    def nop(self) -> None:
        """
        Пустая операция (не нужна).

        :return: None
        """

    def may_be_number(self) -> int:
        """
        Считывание строки, в которой может быть число,
        но может быть и что-нибудь иное.

        :return: Считанное число либо 0, если строку
        не удалось интерпретировать как число.
        """

        result = 0
        try:
            result = int(self.ansi())
        except ValueError:
            pass
        return result

    def number(self) -> int:
        """
        Считывание строки, в которой ожидается число (возможно,
        со знаком) в десятеричной системе счисления.

        :return: Считанное число
        """
        # noinspection PyTypeChecker
        return int(self.read())  # type: ignore

    def read(self) -> memoryview:
        """
        Считываем строку в сыром виде.

        :return: memoryview на сырые байты строки.
        """

        length = len(self._memory)
        start = self._pos
        if self._pos >= length:
            return self._view[0:0]

        while True:
            position = self._memory.find(0x0D, start)
            if position < 0:
                result = self._view[self._pos :]
                self._pos = length
                return result

            plus1 = position + 1
            if plus1 < length:
                if self._memory[plus1] == 0x0A:
                    result = self._view[self._pos : position]
                    self._pos = plus1 + 1
                    return result
                start = plus1

    def utf(self) -> str:
        """
        Считывание одной строки в кодировке UTF-8.

        :return: Строка (возможно, пустая)
        """
        # noinspection PyTypeChecker
        return str(self.read(), UTF)  # type: ignore

    def utf_n(self, count: int) -> "List[str]":
        """
        Считывание не менее указанного количества строк
        в кодировке UTF-8.

        :param count: Сколько строк надо
        :return: Список строк или None
        """
        result = []
        for _ in range(count):
            line = self.utf()
            if not line:
                return []
            result.append(line)
        return result

    def utf_remaining_text(self) -> str:
        """
        Получение всего оставшегося текста ответа сервера
        как строки в кодировке UTF-8.

        :return: Строка (возможно, пустая)
        """
        # noinspection PyTypeChecker
        return str(self._view[self._pos :], UTF)  # type: ignore

    def utf_remaining_lines(self) -> "List[str]":
        """
        Получение всего оставшегося текста ответа сервера
        как списка строк в кодировке UTF-8.

        :return: Список строк (возможно, пустой)
        """

        result = []
        while True:
            line = self.utf()
            if not line:
                break
            result.append(line)
        return result

    def __str__(self):
        return str(self.return_code)

    def __enter__(self):
        # Пока не знаю, что делать
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return exc_type is None


__all__ = ["ServerResponse"]



# coding: utf-8

"""
Серверные процессы.
"""


class Process:
    """
    Информация о запущенном на сервере процессе.
    """

    __slots__ = ('number', 'ip_address', 'name', 'client_id', 'workstation',
                 'started', 'last_command', 'command_number',
                 'process_id', 'state')

    def __init__(self) -> None:
        self.number: str = ''
        self.ip_address: str = ''
        self.name: str = ''
        self.client_id: str = ''
        self.workstation: str = ''
        self.started: str = ''
        self.last_command: str = ''
        self.command_number: str = ''
        self.process_id: str = ''
        self.state: str = ''

    def __str__(self):
        buffer = [self.number, self.ip_address, self.name, self.client_id,
                  self.workstation, self.started, self.last_command,
                  self.command_number, self.process_id, self.state]
        return '\n'.join(x for x in buffer if x)


__all__ = ['Process']



# coding: utf-8

"""
Ошибка, специфичная для ИРБИС.
"""

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Union
    from cbsvibpyirbis.specification import FileSpecification


def get_error_description(code: int) -> str:
    """
    Получение описания ошибки.

    :param code: Код ошибки
    :return: Описание ошибки
    """

    errors = {
        -100: 'Заданный MFN вне пределов БД',
        -101: 'Ошибочный размер полки',
        -102: 'Ошибочный номер полки',
        -140: 'MFN вне пределов БД',
        -141: 'Ошибка чтения',
        -200: 'Указанное поле отсутствует',
        -201: 'Предыдущая версия записи отсутствует',
        -202: 'Заданный термин не найден (термин не существует)',
        -203: 'Последний термин в списке',
        -204: 'Первый термин в списке',
        -300: 'База данных монопольно заблокирована',
        -301: 'База данных монопольно заблокирована',
        -400: 'Ошибка при открытии файлов MST или XRF (ошибка файла данных)',
        -401: 'Ошибка при открытии файлов IFP (ошибка файла индекса)',
        -402: 'Ошибка при записи',
        -403: 'Ошибка при актуализации',
        -600: 'Запись логически удалена',
        -601: 'Запись физически удалена',
        -602: 'Запись заблокирована на ввод',
        -603: 'Запись логически удалена',
        -605: 'Запись физически удалена',
        -607: 'Ошибка autoin.gbl',
        -608: 'Ошибка версии записи',
        -700: 'Ошибка создания резервной копии',
        -701: 'Ошибка восстановления из резервной копии',
        -702: 'Ошибка сортировки',
        -703: 'Ошибочный термин',
        -704: 'Ошибка создания словаря',
        -705: 'Ошибка загрузки словаря',
        -800: 'Ошибка в параметрах глобальной корректировки',
        -801: 'ERR_GBL_REP',
        -801: 'ERR_GBL_MET',
        -1111: 'Ошибка исполнения сервера (SERVER_EXECUTE_ERROR)',
        -2222: 'Ошибка в протоколе (WRONG_PROTOCOL)',
        -3333: 'Незарегистрированный клиент (ошибка входа на сервер) ' +
               '(клиент не в списке)',
        -3334: 'Клиент не выполнил вход на сервер (клиент не используется)',
        -3335: 'Неправильный уникальный идентификатор клиента',
        -3336: 'Нет доступа к командам АРМ',
        -3337: 'Клиент уже зарегистрирован',
        -3338: 'Недопустимый клиент',
        -4444: 'Неверный пароль',
        -5555: 'Файл не существует',
        -6666: 'Сервер перегружен. Достигнуто максимальное число ' +
               'потоков обработки',
        -7777: 'Не удалось запустить/прервать поток администратора ' +
               '(ошибка процесса)',
        -8888: 'Общая ошибка',
    }

    if code >= 0:
        return 'Нормальное завершение'

    if code not in errors:
        return 'Неизвестная ошибка'

    return errors[code]


class IrbisError(Exception):
    """
    Исключнение - ошибка протокола.
    """

    __slots__ = ('code', 'message')

    def __init__(self, code: 'Union[int, str]' = 0) -> None:
        super().__init__(self)
        if isinstance(code, int):
            self.code: int = code
        if isinstance(code, str):
            self.message: str = code

    def __str__(self):
        if self.message is not None:
            return self.message
        return f'{self.code}: {get_error_description(self.code)}'


class IrbisFileNotFoundError(IrbisError):
    """
    Файл на сервере не найден.
    """

    __slots__ = ('filename',)

    def __init__(self, filename: 'Union[str, FileSpecification]') -> None:
        super().__init__()
        self.filename: str = str(filename)

    def __str__(self):
        return f'File not found: {self.filename}'


__all__ = ['IrbisError', 'IrbisFileNotFoundError']



# coding: utf-8

"""
Работа с OPT-файлами.
"""

import re
from typing import TYPE_CHECKING
from cbsvibpyirbis._common import ANSI, STOP_MARKER
from cbsvibpyirbis.saveable import Saveable
if TYPE_CHECKING:
    from typing import List, Optional


class OptLine:
    """
    Строка в OPT-файле.
    """

    __slots__ = 'pattern', 'worksheet'

    def __init__(self, pattern: str = '', worksheet: str = '') -> None:
        self.pattern: str = pattern
        self.worksheet: str = worksheet

    def parse(self, text: str) -> None:
        """
        Parse the line.

        :param text: Text to parse
        :return: None
        """
        parts = re.split(r'\s+', text.strip())
        self.pattern = parts[0]
        self.worksheet = parts[1]


class OptFile(Saveable):
    """
    OPT-файл.
    """

    WILDCARD = '+'

    __slots__ = 'lines', 'length', 'tag'

    def __init__(self):
        self.lines: 'List[OptLine]' = []
        self.length: int = 5
        self.tag: int = 920

    def parse(self, text: 'List[str]') -> None:
        """
        Parse the text for OPT table.

        :param text: Text to parse
        :return: None
        """
        self.tag = int(text[0])
        self.length = int(text[1])
        for line in text[2:]:
            if not line:
                continue
            line = line.strip()
            if not line:
                continue
            if line.startswith('*'):
                continue
            one = OptLine()
            one.parse(line)
            self.lines.append(one)

    @staticmethod
    def same_char(pattern: str, testable: str) -> bool:
        """
        Compare the character against the pattern.

        :param pattern: Pattern character
        :param testable: Character to examine
        :return: True or False
        """
        if pattern == OptFile.WILDCARD:
            return True
        return pattern.lower() == testable.lower()

    def same_text(self, pattern: str, testable: str) -> bool:
        """
        Compare tag value against the OPT pattern.

        :param pattern: Pattern to use
        :param testable: Tag value to examine
        :return: True or False
        """

        if not pattern:
            return False

        if not testable:
            return pattern[0] == OptFile.WILDCARD

        pattern_index = 0
        testable_index = 0

        while True:
            pattern_char = pattern[pattern_index]
            testable_char = testable[testable_index]
            pattern_index += 1
            testable_index += 1
            pattern_next = pattern_index < len(pattern)
            testable_next = testable_index < len(testable)

            if pattern_next and not testable_next:
                if pattern_char == OptFile.WILDCARD:
                    while pattern_index < len(pattern):
                        pattern_char = pattern[pattern_index]
                        pattern_index += 1

                        if pattern_char != OptFile.WILDCARD:
                            return False

                    return True

            if pattern_next != testable_next:
                return False

            if not pattern_next:
                return True

            if not self.same_char(pattern_char, testable_char):
                return False

    def resolve_worksheet(self, tag: str) -> 'Optional[str]':
        """
        Resolve worksheet for the specified tag value.

        :param tag: Tag value, e. g. "SPEC"
        :return: Worksheet name or None
        """

        for line in self.lines:
            if self.same_text(line.pattern, tag):
                return line.worksheet

        return None

    def __str__(self):
        result = [str(self.tag), str(self.length)]
        for line in self.lines:
            result.append(line.pattern.ljust(6) + line.worksheet)
        result.append(STOP_MARKER)
        return '\n'.join(result)


def load_opt_file(filename: str) -> OptFile:
    """
    Load the OPT from the specified file.

    :param filename: Name of the file
    :return: OPT file
    """

    result = OptFile()
    with open(filename, 'rt', encoding=ANSI) as stream:
        lines = stream.readlines()
        result.parse(lines)
    return result


__all__ = ['load_opt_file', 'OptFile', 'OptLine']



# coding: utf-8

"""
Всё, связанное с INI-файлами.
"""

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Iterable, List, Optional, Union


class IniLine:
    """
    Строка INI-файла.
    """

    __slots__ = 'key', 'value'

    def __init__(self, key: 'Optional[str]' = None,
                 value: 'Optional[str]' = None) -> None:
        self.key = key
        self.value = value

    def __str__(self):
        return str(self.key) + '=' + str(self.value)

    def __repr__(self):
        return self.__str__()


def same_key(first: 'Optional[str]', second: 'Optional[str]') -> bool:
    """
    Сравнение двух ключей (с точностью до регистра).

    :param first: Первый ключ (может отсутствовать)
    :param second: Второй ключ (может отсутствовать)
    :return: True при совпадении
    """

    if not first or not second:
        return False
    return first.upper() == second.upper()


class IniSection:
    """
    Секция INI-файла.
    """

    __slots__ = 'name', 'lines'

    def __init__(self, name: 'Optional[str]' = None) -> None:
        self.name: 'Optional[str]' = name
        self.lines: 'List[IniLine]' = []

    def find(self, key: str) -> 'Optional[IniLine]':
        """
        Нахождение строки с указанным ключом.

        :param key: Искомый ключ
        :return: Найденная строка или None
        """
        for line in self.lines:
            if same_key(line.key, key):
                return line
        return None

    def get_value(self, key: str,
                  default: 'Optional[str]' = None) -> 'Optional[str]':
        """
        Получение значения строки с указанным ключом.

        :param key: Искомый ключ
        :param default: Значение по умолчанию
        :return: Найденное значение или значение по умолчанию
        """
        found = self.find(key)
        return found.value if found else default

    def set_value(self, key: str, value: str) -> None:
        """
        Установка значения строки с указанным ключом.

        :param key: Искомый ключ
        :param value: Устанавливаемое значение
        :return: None
        """
        found = self.find(key)
        if found:
            found.value = value
        else:
            found = IniLine(key, value)
            self.lines.append(found)

    def remove(self, key: str) -> None:
        """
        Удаление строки с указанным ключом.

        :param key: Искомый ключ
        :return: None
        """
        found = self.find(key)
        if found:
            self.lines.remove(found)

    def __str__(self):
        result = []
        if self.name:
            result.append('[' + self.name + ']')
        for line in self.lines:
            result.append(str(line))
        return '\n'.join(result)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        yield from self.lines

    def __getitem__(self, item: 'Union[str, int]'):
        if isinstance(item, int):
            return self.lines[item]
        return self.get_value(item)

    def __len__(self):
        return len(self.lines)

    def __bool__(self):
        return bool(self.lines)


class IniFile:
    """
    INI-файл.
    """

    __slots__ = ('sections',)

    def __init__(self):
        self.sections = []

    def find(self, name: str) -> 'Optional[IniSection]':
        """
        Нахождение секции с указанным именем.

        :param name: Имя секции
        :return: Найденная секция или None
        """
        for section in self.sections:
            if not name and not section.name:
                return section
            if same_key(section.name, name):
                return section
        return None

    def get_or_create(self, name: str) -> IniSection:
        """
        Получение или создание при отсутствии секции с указанным именем.

        :param name: Имя секции
        :return: Найденная или созданная секция
        """
        result = self.find(name)
        if not result:
            result = IniSection(name)
            self.sections.append(result)
        return result

    def get_value(self, name: str, key: str,
                  default: 'Optional[str]' = None) -> 'Optional[str]':
        """
        Получение значения строки с указанными именем секции и ключом.

        :param name: Имя секции
        :param key: Ключ строки в секции
        :param default: Значение по умолчанию
        :return: Найденное значение или значение по умолчанию
        """
        section = self.find(name)
        if section:
            return section.get_value(key, default)
        return default

    def set_value(self, name: str, key: str, value: str) -> None:
        """
        Установка значения строки с указанными именем секции и ключом.

        :param name: Имя секции
        :param key: Ключ строки в секции
        :param value: Устанавливаемое значение
        :return: None
        """
        section = self.get_or_create(name)
        if not section:
            section = IniSection(name)
            self.sections.append(section)
        section.set_value(key, value)

    def parse(self, text: 'Iterable[str]') -> None:
        """
        Разбор текстового представления INI-файла.

        :param text: Последовательность текстовых строк
        :return: None
        """
        section = None
        for line in text:
            line = line.strip()
            if not line:
                continue
            if line.startswith('['):
                name = line[1:-1]
                section = self.get_or_create(name)
            else:
                parts = line.split('=', 1)
                key = parts[0]
                value = None
                if len(parts) > 1:
                    value = parts[1]
                item = IniLine(key, value)
                if section is None:
                    section = IniSection()
                    self.sections.append(section)
                section.lines.append(item)

    def __str__(self):
        result = []
        for section in self.sections:
            result.append(str(section))
        return '\n'.join(result)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        yield from self.sections

    def __getitem__(self, item: 'Union[str, int]'):
        if isinstance(item, int):
            return self.sections[item]
        return self.find(item)

    def __len__(self):
        return len(self.sections)

    def __bool__(self):
        return bool(self.sections)


__all__ = ['IniFile', 'IniLine', 'IniSection']



# coding: utf-8

"""
Пользователи системы.
"""

from typing import TYPE_CHECKING
from cbsvibpyirbis._common import safe_str, same_string
from cbsvibpyirbis.response import ServerResponse
if TYPE_CHECKING:
    from typing import List, Optional


class UserInfo:
    """
    Информация о зарегистрированном пользователе системы.
    """

    __slots__ = ('number', 'name', 'password', 'cataloger',
                 'reader', 'circulation', 'acquisitions',
                 'provision', 'administrator')

    def __init__(self, name: 'Optional[str]' = None,
                 password: 'Optional[str]' = None) -> None:
        self.number: 'Optional[str]' = None
        self.name: 'Optional[str]' = name
        self.password: 'Optional[str]' = password
        self.cataloger: 'Optional[str]' = None
        self.reader: 'Optional[str]' = None
        self.circulation: 'Optional[str]' = None
        self.acquisitions: 'Optional[str]' = None
        self.provision: 'Optional[str]' = None
        self.administrator: 'Optional[str]' = None

    # noinspection DuplicatedCode
    @staticmethod
    def parse(response: ServerResponse) -> 'List[UserInfo]':
        """
        Parse the server response for the user info.

        :param response: Response to parse.
        :return: List of user infos.
        """

        result: 'List' = []
        user_count = response.number()
        lines_per_user = response.number()
        if not user_count or not lines_per_user:
            return result
        for _ in range(user_count):
            user: UserInfo = UserInfo()
            user.number = response.ansi()
            user.name = response.ansi()
            user.password = response.ansi()
            user.cataloger = response.ansi()
            user.reader = response.ansi()
            user.circulation = response.ansi()
            user.acquisitions = response.ansi()
            user.provision = response.ansi()
            user.administrator = response.ansi()
            result.append(user)
        return result

    @staticmethod
    def format_pair(prefix: str, value: str, default: str) -> str:
        """
        Format the pair prefix=value.

        :param prefix: Prefix to use
        :param value: Value to use
        :param default: Default value
        :return: Formatted text
        """

        if same_string(value, default):
            return ''

        return prefix + '=' + safe_str(value) + ';'

    def encode(self):
        """
        Encode the user info.

        :return: Text representation of the user info.
        """

        return self.name + '\n' + self.password + '\n' \
            + UserInfo.format_pair('C', self.cataloger, "irbisc.ini") \
            + UserInfo.format_pair('R', self.reader, "irbisr.ini") \
            + UserInfo.format_pair('B', self.circulation, "irbisb.ini") \
            + UserInfo.format_pair('M', self.acquisitions, "irbism.ini") \
            + UserInfo.format_pair('K', self.provision, "irbisk.ini") \
            + UserInfo.format_pair('A', self.administrator, "irbisa.ini")

    def __str__(self):
        buffer = [self.number, self.name, self.password, self.cataloger,
                  self.reader, self.circulation, self.acquisitions,
                  self.provision, self.administrator]
        return ' '.join(x for x in buffer if x)


__all__ = ['UserInfo']



# coding: utf-8

"""
Работа с TRE-файлами.
"""

from typing import TYPE_CHECKING
from cbsvibpyirbis._common import ANSI, safe_str
from cbsvibpyirbis.saveable import Saveable
if TYPE_CHECKING:
    from typing import Iterable, List, Optional


class TreeNode:
    """
    TRE-file line.
    """

    __slots__ = 'children', 'value', 'level'

    def __init__(self, value: 'Optional[str]' = None, level: int = 0) -> None:
        self.children: 'List' = []
        self.value: 'Optional[str]' = value
        self.level: int = level

    def add(self, name: str):
        """
        Add a child node.

        :param name: Name of the child.
        :return: Added subnode
        """

        result = TreeNode(name)
        self.children.append(result)
        return result

    def write(self) -> 'List[str]':
        """
        Represent the node and its child as lines.

        :return: List of lines
        """
        result = [TreeFile.INDENT * self.level + safe_str(self.value)]
        for child in self.children:
            inner = child.write()
            result.extend(inner)
        return result

    def __str__(self):
        return self.value


class TreeFile(Saveable):
    """
    TRE-file.
    """

    INDENT = '\u0009'

    __slots__ = ('roots',)

    def __init__(self):
        self.roots: 'List[TreeNode]' = []

    @staticmethod
    def _count_indent(text: str) -> int:
        result = 0
        for char in text:
            if char == TreeFile.INDENT:
                result += 1
            else:
                break
        return result

    @staticmethod
    def _arrange_level(nodes: 'List[TreeNode]', level: int) -> None:
        count = len(nodes)
        index = 0
        while index < count:
            index = TreeFile._arrange_nodes(nodes, level, index, count)

    @staticmethod
    def _arrange_nodes(nodes: 'List[TreeNode]', level: int,
                       index: int, count: int) -> int:
        nxt = index + 1
        level2 = level + 1
        parent = nodes[index]
        while nxt < count:
            child = nodes[nxt]
            if child.level <= level:
                break
            if child.level == level2:
                parent.children.append(child)
            nxt += 1
        return nxt

    def add(self, name: str) -> TreeNode:
        """
        Add zero level node with specified name.

        :param name: Name of the node
        :return: Created node
        """

        result = TreeNode(name)
        self.roots.append(result)
        return result

    @staticmethod
    def determine_level(nodes: 'Iterable[TreeNode]', current: int) -> None:
        """
        Determine level of the nodes.

        :param nodes: Nodes to process
        :param current: Current level
        :return: None
        """

        for node in nodes:
            node.level = current
            TreeFile.determine_level(node.children, current + 1)

    def parse(self, text: 'Iterable[str]') -> None:
        """
        Parse the text for the tree structure.

        :param text: Text to parse
        :return: None
        """

        nodes = []
        for line in text:
            level = TreeFile._count_indent(line)
            line = line[level:]
            node = TreeNode(line, level)
            nodes.append(node)

        max_level = max(node.level for node in nodes)
        for level in range(max_level):
            TreeFile._arrange_level(nodes, level)

        for node in nodes:
            if node.level == 0:
                self.roots.append(node)

    def __str__(self):
        TreeFile.determine_level(self.roots, 0)
        result = []
        for node in self.roots:
            result.extend(node.write())
        return '\n'.join(result)


def load_tree_file(filename: str) -> TreeFile:
    """
    Load the tree from the specified file.

    :param filename: Name of the file.
    :return: Tree structure
    """

    result = TreeFile()
    with open(filename, 'rt', encoding=ANSI) as stream:
        lines = stream.readlines()
        result.parse(lines)
    return result


__all__ = ['load_tree_file', 'TreeFile', 'TreeNode']



# coding: utf-8

"""
Search expression builder.
"""

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any

###############################################################################


class Search:
    """
    Search expression builder class.
    """

    __slots__ = ('buffer',)

    def __init__(self) -> None:
        self.buffer = ''

    @staticmethod
    def all() -> 'Search':
        """
        All documents in the database.
        """
        result = Search()
        result.buffer = 'I=$'
        return result

    def and_(self, *items: 'Any') -> 'Search':
        """
        Logical AND.
        """
        self.buffer = '(' + self.buffer
        for item in items:
            self.buffer = self.buffer + ' * ' + Search.wrap(item)
        self.buffer = self.buffer + ')'
        return self

    @staticmethod
    def equals(prefix: str, *values: 'Any') -> 'Search':
        """
        Prefix search for matching records.
        """
        result = Search()
        text = Search.wrap(prefix + str(values[0]))
        if len(values) > 1:
            text = '(' + text
            for item in values[1:len(values)]:
                text = text + ' + ' + Search.wrap(prefix + str(item))
            text = text + ')'
        result.buffer = text
        return result

    @staticmethod
    def need_wrap(text: str) -> bool:
        """
        Need wrap the text?
        """
        if not text:
            return True
        if text[0] in '"(':
            return False
        for one in text:
            if one in ' +*^()"#':
                return True
        return False

    def not_(self, text: 'Any') -> 'Search':
        """
        Logical NOT.
        """
        self.buffer = '(' + self.buffer + ' ^ ' + Search.wrap(text) + ')'
        return self

    def or_(self, *items: 'Any') -> 'Search':
        """
        Logical OR.
        """
        self.buffer = '(' + self.buffer
        for item in items:
            self.buffer = self.buffer + ' + ' + Search.wrap(item)
        self.buffer = self.buffer + ')'
        return self

    def same_field(self, *items: 'Any') -> 'Search':
        """
        Logical "Same field"
        """
        self.buffer = '(' + self.buffer
        for item in items:
            self.buffer = self.buffer + ' (G) ' + Search.wrap(item)
        self.buffer = self.buffer + ')'
        return self

    def same_repeat(self, *items: 'Any') -> 'Search':
        """
        Logical "Same field repeat"
        """
        self.buffer = '(' + self.buffer
        for item in items:
            self.buffer = self.buffer + ' (F) ' + Search.wrap(item)
        self.buffer = self.buffer + ')'
        return self

    @staticmethod
    def wrap(text: 'Any') -> str:
        """
        Wrap the text if needed.
        """
        text = str(text)
        if Search.need_wrap(text):
            return '"' + text + '"'
        return text

    def __str__(self):
        return self.buffer

###############################################################################


def keyword(*values: 'Any') -> Search:
    """
    Поиск по ключевым словам.

    :param values: Искомые ключевые слова.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('K=', *values)


def author(*values: 'Any') -> Search:
    """
    Поиск по индивидуальным авторам.

    :param values: Искомые авторы.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('A=', *values)


def title(*values: 'Any') -> Search:
    """
    Поиск по заглавиям.

    :param values: Искомые заглавия.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('T=', *values)


def number(*values: 'Any') -> Search:
    """
    Поиск по инвентарным номерам.

    :param values: Искомые номера.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('IN=', *values)


def publisher(*values: 'Any') -> Search:
    """
    Поиск по издательству.

    :param values: Искомые издательства.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('O=', *values)


def place(*values: 'Any') -> Search:
    """
    Поиск по месту издания (городу).

    :param values: Искомые места (города).
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('MI=', *values)


def subject(*values: 'Any') -> Search:
    """
    Поиск по предметным рубрикам.

    :param values: Искомые рубрики.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('S=', *values)


def language(*values: 'Any') -> Search:
    """
    Поиск по языку документа.

    :param values: Искомые языки.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('J=', *values)


def year(*values: 'Any') -> Search:
    """
    Поиск по году издания.

    :param values: Искомые годы.
    :return: Построитель запросов.
    """
    assert values
    return Search.equals('G=', *values)


def magazine(*values: 'Any') -> Search:
    """
    Поиск по заглавиям журналов.

    :param values: Искомые заглавия журналов.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('TJ=', *values)


def document_kind(*values: 'Any') -> Search:
    """
    Поиск по виду документа.

    :param values: Искомые виды документов.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('V=', *values)


def udc(*values: 'Any') -> Search:
    """
    Поиск по индексам УДК.

    :param values: Искомые индексы ББК.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('U=', *values)


def bbk(*values: 'Any') -> Search:
    """
    Поиск по индексам ББК.

    :param values: Искомые индексы ББК.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('BBK=', *values)


def rzn(*values: 'Any') -> Search:
    """
    Поиск по разделу знаний.

    :param values: Искомые разделы знаний.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('RZN=', *values)


def mhr(*values: 'Any') -> Search:
    """
    Поиск по месту хранения.

    :param values: Искомые места хранения.
    :return: Построитель запросов.
    """
    assert values

    return Search.equals('MHR=', *values)



# coding: utf-8

"""
Всё, связанное с поиском.
"""

from typing import TYPE_CHECKING
from cbsvibpyirbis._common import safe_int, safe_str
from cbsvibpyirbis.ini import IniFile
if TYPE_CHECKING:
    from typing import Any, List, Optional


class FoundLine:
    """
    Found item in search answer.
    """

    __slots__ = ('mfn', 'description')

    def __init__(self) -> None:
        self.mfn: int = 0
        self.description: 'Optional[str]' = None

    def parse_line(self, line: str) -> None:
        """
        Разбор серверного представления найденной записи.

        :param line: Строка с найденной записью.
        :return: None.
        """
        parts = line.split('#', 2)
        self.mfn = safe_int(parts[0])
        if len(parts) > 1:
            self.description = parts[1]

    def __str__(self):
        return f"{self.mfn}#{self.description}"


class SearchParameters:
    """
    Parameters for search request.
    """

    __slots__ = ('database', 'first', 'format', 'max_mfn', 'min_mfn',
                 'number', 'expression', 'sequential', 'filter', 'utf')

    def __init__(self, expression: 'Optional[str]' = None,
                 number: int = 0) -> None:
        self.database: 'Optional[str]' = None
        self.first: int = 1
        self.format: 'Optional[str]' = None
        self.max_mfn: int = 0
        self.min_mfn: int = 0
        self.number: int = number
        self.expression = expression
        self.sequential: 'Optional[str]' = None
        self.filter: 'Optional[str]' = None
        self.utf = False

    def encode(self, query: 'Any', connection: 'Any') -> None:
        """
        Кодирование параметров поиска в клиентский запрос.

        :param query: Клиентский запрос.
        :param connection: Подключение.
        :return: None.
        """
        database = self.database or connection.database
        query.ansi(database)
        query.utf(self.expression)
        query.add(self.number)
        query.add(self.first)
        query.ansi(self.format)
        query.add(self.min_mfn)
        query.add(self.max_mfn)
        query.ansi(self.sequential)

    def __str__(self):
        return self.expression


class SearchScenario:
    """
    Сценарий поиска.
    """

    __slots__ = ('name', 'prefix', 'type', 'menu', 'old',
                 'correction', 'truncation', 'hint',
                 'mod_by_dic_auto', 'logic', 'advance',
                 'format')

    def __init__(self, name: 'Optional[str]' = None) -> None:
        self.name: 'Optional[str]' = name
        self.prefix: 'Optional[str]' = None
        self.type: int = 0
        self.menu: 'Optional[str]' = None
        self.old: 'Optional[str]' = None
        self.correction: 'Optional[str]' = None
        self.truncation: bool = False
        self.hint: 'Optional[str]' = None
        self.mod_by_dic_auto: 'Optional[str]' = None
        self.logic: int = 0
        self.advance: 'Optional[str]' = None
        self.format: 'Optional[str]' = None

    @staticmethod
    def parse(ini: IniFile) -> 'List[SearchScenario]':
        """
        Parse the INI file for the search scenarios.

        :param ini: INI file
        :return: List of search scenarios
        """

        section = ini.find('SEARCH')
        if not section:
            return []
        count = safe_int(safe_str(section.get_value('ItemNumb', '0')))
        if not count:
            return []
        result: 'List[SearchScenario]' = []
        for i in range(count):
            name = section.get_value(f'ItemName{i}')
            scenario = SearchScenario(name)
            result.append(scenario)
            scenario.prefix = section.get_value(f'ItemPref{i}') or ''
            scenario.type = safe_int(safe_str(section.get_value(
                f'ItemDictionType{i}', '0')))
            scenario.menu = section.get_value(f'ItemMenu{i}')
            scenario.old = None
            scenario.correction = section.get_value(f'ItemModByDic{i}')
            scenario.truncation = bool(section.get_value(
                f'ItemTranc{i}', '0'))
            scenario.hint = section.get_value(f'ItemHint{i}')
            scenario.mod_by_dic_auto = section.get_value(
                f'ItemModByDicAuto{i}')
            scenario.logic = safe_int(safe_str(section.get_value(
                f'ItemLogic{i}', '0')))
            scenario.advance = section.get_value(f'ItemAdv{i}')
            scenario.format = section.get_value(f'ItemPft{i}')
        return result

    def __str__(self):
        if not self.prefix:
            return safe_str(self.name)

        return safe_str(self.name) + ' ' + safe_str(self.prefix)


class TextParameters:
    """
    Параметры полнотекстового поиска для ИРБИС64+.
    """
    __slots__ = ('request', 'words', 'morphology', 'prefix',
                 'max_distance', 'context', 'max_count',
                 'cell_type', 'cell_count')

    def __init__(self) -> None:
        self.request: 'Optional[str]' = None
        self.words: 'List[str]' = []
        self.morphology: bool = False
        self.prefix: 'Optional[str]' = "KT="
        self.max_distance: int = -1
        self.context: 'Optional[str]' = None
        self.max_count: int = 100
        self.cell_type: 'Optional[str]' = None
        self.cell_count: int = 5

    def encode(self, query: 'Any') -> None:
        """
        Кодирование параметров полнотекстового поиска в клиентский запрос.

        :param query: Клиентский запрос.
        :return: None
        """
        delimiter: str = '\x1F'
        morpho = '0'
        if self.morphology:
            morpho = '1'
        line = (self.request or '') + delimiter + \
            ','.join(self.words) + delimiter + \
            morpho + delimiter + \
            (self.prefix or '') + delimiter + \
            str(self.max_distance) + delimiter + \
            (self.context or '') + delimiter + \
            str(self.max_count) + delimiter + \
            (self.cell_type or '') + delimiter + \
            str(self.cell_count)
        query.utf(line)


class TextResult:
    """
    Результат полнотекстового поиска
    """
    __slots__ = ('mfn', 'pages', 'formatted')

    def __init__(self) -> None:
        self.formatted: 'Optional[str]' = None
        self.mfn: int = 0
        self.pages: 'List[int]' = []

    def decode(self, line: str) -> None:
        """
        Декодирование строки с одним результатом полнотекстового поиска.
        """
        parts = line.split('#', 2)
        self.mfn = int(parts[0])
        if len(parts) == 3:
            self.formatted = parts[2]
        if len(parts) > 1:
            parts = parts[1].split('\x1F')
            for part in parts:
                if part.isdecimal():
                    page = int(part)
                    self.pages.append(page)

    def __str__(self):
        return str(self.mfn)


class CellResult:
    """
    Фасет, т. е. строка поискового словаря, близко соответствующая
    результатам поиска.
    """
    __slots__ = ('count', 'term')

    def __init__(self):
        self.count: int = 0
        self.term: 'Optional[str]' = None

    def decode(self, line: str) -> None:
        """
        Декодирование строки с одним фасетом.
        """
        parts = line.split('#', 1)
        self.count = int(parts[0])
        self.term = parts[1]

    def __str__(self):
        return self.term or '--nothing--'


__all__ = ['CellResult', 'FoundLine', 'SearchParameters', 'SearchScenario',
           'TextParameters', 'TextResult']



# coding: utf-8

"""
Спецификация файла на сервере ИРБИС64.
"""

from typing import TYPE_CHECKING
from cbsvibpyirbis._common import SYSTEM
if TYPE_CHECKING:
    from typing import Optional


class FileSpecification:
    """
    Путь на файл path.database.filename
    path – код путей:
    0 – общесистемный путь.
    1 – путь размещения сведений о базах данных сервера ИРБИС64
    2 – путь на мастер-файл базы данных.
    3 – путь на словарь базы данных.
    10 – путь на параметрию базы данных.
    database – имя базы данных
    filename – имя требуемого файла с расширением
    В случае чтения ресурса по пути 0 и 1 имя базы данных не задается.

    В общем случае может выглядеть так:

    path.database.@filename
    или
    path.database.&filename&content

    где @ означает двоичный файл
    & означает наличие содержимого и отделяет имя файла от его содержимого

    """

    __slots__ = 'binary', 'path', 'database', 'filename', 'content'

    def __init__(self, path: int, database: 'Optional[str]',
                 filename: str) -> None:
        self.binary: bool = False
        self.path: int = path
        self.database: 'Optional[str]' = database
        self.filename: str = filename
        self.content: 'Optional[str]' = None

    @staticmethod
    def system(filename: str) -> 'FileSpecification':
        """
        Создание спецификации файла, лежащего в системной папке ИРБИС64.

        :param filename: Имя файла
        :return: Спецификация файла
        """
        return FileSpecification(SYSTEM, None, filename)

    @staticmethod
    def parse(text: str) -> 'FileSpecification':
        """
        Разбор текстового представления спецификации.

        :param text: Текст для разбора.
        :return: Спецификация.
        """
        parts = text.split('.', 2)
        result = FileSpecification(int(parts[0]), parts[1], parts[2])
        if result.filename.startswith('@'):
            result.filename = result.filename[1:]
            result.binary = True
            parts = result.filename.split('&', 1)
            result.filename = parts[0]
            result.content = parts[1]
        if result.filename.startswith('&'):
            parts = result.filename[1:].split('&', 1)
            result.filename = parts[0]
            result.content = parts[1]
        return result

    def __str__(self):
        result = self.filename

        if self.binary:
            result = '@' + self.filename
        else:
            if self.content:
                result = '&' + self.filename

        if self.path in (0, 1):
            result = str(self.path) + '..' + result
        else:
            result = str(self.path) + '.' + self.database + '.' + result

        if self.content:
            result = result + '&' + str(self.content)

        return result


__all__ = ['FileSpecification']



# coding: utf-8

"""
Модуль содержит основную функциональность по работе с сервером ИРБИС64,
в т. ч. для манипуляций с записями.
"""

# ManagedIrbis ported from C# to Python.
# Python supported: 3.6 and higher
# IRBIS64 supported: 2014 and higher

from cbsvibpyirbis._common import ADMINISTRATOR, BRIEF, CATALOGER, close_async, \
    init_async, get_event_loop, LAST, LOCKED, LOGICALLY_DELETED, \
    NON_ACTUALIZED, NOT_CONNECTED, PHYSICALLY_DELETED, prepare_format, \
    remove_comments

from cbsvibpyirbis.alphabet import AlphabetTable, load_alphabet_table, \
    UpperCaseTable, load_uppercase_table
from cbsvibpyirbis.database import DatabaseInfo
from cbsvibpyirbis.direct import DirectAccess, InvertedFile, MstControl, MstField,\
    MstFile, MstEntry, MstLeader, MstRecord, XrfFile, XrfRecord
from cbsvibpyirbis.error import IrbisError, IrbisFileNotFoundError
from cbsvibpyirbis.export import read_iso_record, read_text_record, STOP_MARKER, \
    write_iso_record, write_text_record
from cbsvibpyirbis.ini import IniFile, IniLine, IniSection
from cbsvibpyirbis.menus import load_menu, MenuEntry, MenuFile
from cbsvibpyirbis.opt import load_opt_file, OptFile
from cbsvibpyirbis.par import load_par_file, ParFile
from cbsvibpyirbis.process import Process
from cbsvibpyirbis.query import ClientQuery
from cbsvibpyirbis.records import Field, RawRecord, Record, SubField
from cbsvibpyirbis.resource import Resource, ResourceDictionary
from cbsvibpyirbis.response import ServerResponse
from cbsvibpyirbis.search import CellResult, FoundLine, SearchParameters, \
    SearchScenario, TextParameters, TextResult
from cbsvibpyirbis.specification import FileSpecification
from cbsvibpyirbis.stats import ClientInfo, ServerStat
from cbsvibpyirbis.table import TableDefinition
from cbsvibpyirbis.terms import PostingParameters, TermInfo, TermParameters, \
    TermPosting
from cbsvibpyirbis.tree import load_tree_file, TreeFile, TreeNode
from cbsvibpyirbis.user import UserInfo
from cbsvibpyirbis.version import ServerVersion

from cbsvibpyirbis.connection import Connection


__version__ = '0.2'
__author__ = 'Alexey Mironov'
__email__ = 'amironov73@gmail.com'

__title__ = 'irbis'
__summary__ = 'universal client software for IRBIS64 - popular ' + \
              'russian library automation system'
__uri__ = 'http://arsmagna.ru'

__license__ = 'MIT License'
__copyright__ = 'Copyright 2018-2021 Alexey Mironov'

__all__ = ['ADMINISTRATOR', 'AlphabetTable', 'BRIEF', 'CATALOGER',
           'CellResult', 'ClientInfo', 'ClientQuery', 'close_async',
           'Connection', 'DatabaseInfo', 'DirectAccess', 'irbis_event_loop',
           'IrbisError', 'IrbisFileNotFoundError', 'Field',
           'FileSpecification', 'FoundLine', 'IniFile', 'IniLine',
           'IniSection', 'init_async', 'InvertedFile',
           'load_alphabet_table', 'load_menu', 'load_opt_file',
           'load_par_file', 'load_tree_file', 'load_uppercase_table',
           'LAST', 'LOCKED', 'LOGICALLY_DELETED', 'MenuEntry',
           'MenuFile', 'MstControl', 'MstField', 'MstFile',
           'MstEntry', 'MstLeader', 'MstRecord', 'NON_ACTUALIZED',
           'NOT_CONNECTED', 'OptFile', 'ParFile', 'PHYSICALLY_DELETED',
           'PostingParameters', 'prepare_format', 'Process', 'RawRecord',
           'read_iso_record', 'read_text_record', 'Record', 'remove_comments',
           'Resource', 'ResourceDictionary', 'SearchParameters',
           'SearchScenario', 'ServerResponse', 'ServerStat', 'ServerVersion',
           'STOP_MARKER', 'SubField', 'TableDefinition', 'TermInfo',
           'TermParameters', 'TextResult', 'TermPosting', 'TextParameters',
           'TreeFile', 'TreeNode', 'UpperCaseTable', 'UserInfo',
           'write_iso_record', 'write_text_record', 'XrfFile', 'XrfRecord']
